/*
    Android Asynchronous Http Client
    Copyright (c) 2011 James Smith <james@loopj.com>
    http://loopj.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
 */

package com.chinanetcenter.wcs.android.http;

import com.chinanetcenter.wcs.android.utils.WCSLogUtil;

import org.apache.http.HttpEntity;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.utils.URLEncodedUtils;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.protocol.HTTP;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A collection of string request parameters or files to send along with
 * requests made from an {@link com.chinanetcenter.wcs.android.http.AsyncHttpClient} instance.
 * <p>
 * &nbsp;
 * </p>
 * For example:
 * <p>
 * &nbsp;
 * </p>
 * 
 * <pre>
 * SimpleRequestParams params = new SimpleRequestParams();
 * params.put(&quot;username&quot;, &quot;james&quot;);
 * params.put(&quot;password&quot;, &quot;123456&quot;);
 * params.put(&quot;email&quot;, &quot;my@email.com&quot;);
 * params.put(&quot;profile_picture&quot;, new File(&quot;pic.jpg&quot;)); // Upload a File
 * params.put(&quot;profile_picture2&quot;, someInputStream); // Upload an InputStream
 * params.put(&quot;profile_picture3&quot;, new ByteArrayInputStream(someBytes)); // Upload
 * // some
 * // bytes
 * 
 * Map&lt;String, String&gt; map = new HashMap&lt;String, String&gt;();
 * map.put(&quot;first_name&quot;, &quot;James&quot;);
 * map.put(&quot;last_name&quot;, &quot;Smith&quot;);
 * params.put(&quot;user&quot;, map); // url params:
 * // &quot;user[first_name]=James&amp;user[last_name]=Smith&quot;
 * 
 * Set&lt;String&gt; set = new HashSet&lt;String&gt;(); // unordered collection
 * set.add(&quot;music&quot;);
 * set.add(&quot;art&quot;);
 * params.put(&quot;like&quot;, set); // url params: &quot;like=music&amp;like=art&quot;
 * 
 * List&lt;String&gt; list = new ArrayList&lt;String&gt;(); // Ordered collection
 * list.add(&quot;Java&quot;);
 * list.add(&quot;C&quot;);
 * params.put(&quot;languages&quot;, list); // url params: &quot;languages[]=Java&amp;languages[]=C&quot;
 * 
 * String[] colors = {
 *         &quot;blue&quot;, &quot;yellow&quot;
 * }; // Ordered collection
 * params.put(&quot;colors&quot;, colors); // url params: &quot;colors[]=blue&amp;colors[]=yellow&quot;
 * 
 * List&lt;Map&lt;String, String&gt;&gt; listOfMaps = new ArrayList&lt;Map&lt;String, String&gt;&gt;();
 * Map&lt;String, String&gt; user1 = new HashMap&lt;String, String&gt;();
 * user1.put(&quot;age&quot;, &quot;30&quot;);
 * user1.put(&quot;gender&quot;, &quot;male&quot;);
 * Map&lt;String, String&gt; user2 = new HashMap&lt;String, String&gt;();
 * user2.put(&quot;age&quot;, &quot;25&quot;);
 * user2.put(&quot;gender&quot;, &quot;female&quot;);
 * listOfMaps.add(user1);
 * listOfMaps.add(user2);
 * params.put(&quot;users&quot;, listOfMaps); // url params:
 * // &quot;users[][age]=30&amp;users[][gender]=male&amp;users[][age]=25&amp;users[][gender]=female&quot;
 * 
 * AsyncHttpClient client = new AsyncHttpClient();
 * client.post(&quot;http://myendpoint.com&quot;, params, responseHandler);
 * </pre>
 */
public class SimpleRequestParams implements RequestParams{

	protected final static String LOG_TAG = "SimpleRequestParams";

	protected boolean isRepeatable;

	// protected boolean useJsonStreamer;
	protected ConcurrentHashMap<String, String> urlParams;

	protected ConcurrentHashMap<String, StreamWrapper> streamParams;

	protected ConcurrentHashMap<String, FileWrapper> fileParams;

	protected ConcurrentHashMap<String, Object> urlParamsWithObjects;

	protected String contentEncoding = HTTP.UTF_8;

	/**
	 * Sets content encoding for return value of {@link #getParamString()} and
	 * {@link #createFormEntity()}
	 * <p>
	 * &nbsp;
	 * </p>
	 * Default encoding is "UTF-8"
	 * 
	 * @param encoding String constant from
	 *            {@link org.apache.http.protocol.HTTP}
	 */
	public void setContentEncoding(final String encoding) {
		if (encoding != null)
			this.contentEncoding = encoding;
		else
			WCSLogUtil.d("setContentEncoding called with null attribute");
	}

	/**
	 * Constructs a new empty {@code SimpleRequestParams} instance.
	 */
	public SimpleRequestParams() {
		this((Map<String, String>) null);
	}

	/**
	 * Constructs a new SimpleRequestParams instance containing the key/value string
	 * params from the specified map.
	 * 
	 * @param source the source key/value string map to add.
	 */
	public SimpleRequestParams(Map<String, String> source) {
		init();
		if (source != null) {
			for (Map.Entry<String, String> entry : source.entrySet()) {
				put(entry.getKey(), entry.getValue());
			}
		}
	}

	/**
	 * Constructs a new SimpleRequestParams instance and populate it with a single
	 * initial key/value string param.
	 * 
	 * @param key the key name for the intial param.
	 * @param value the value string for the initial param.
	 */
	public SimpleRequestParams(final String key, final String value) {
		this(new HashMap<String, String>() {
			{
				put(key, value);
			}
		});
	}

	/**
	 * Constructs a new SimpleRequestParams instance and populate it with multiple
	 * initial key/value string param.
	 * 
	 * @param keysAndValues a sequence of keys and values. Objects are
	 *            automatically converted to Strings (including the value
	 *            {@code null}).
	 * @throws IllegalArgumentException if the number of arguments isn't even.
	 */
	public SimpleRequestParams(Object... keysAndValues) {
		init();
		int len = keysAndValues.length;
		if (len % 2 != 0)
			throw new IllegalArgumentException("Supplied arguments must be even");
		for (int i = 0; i < len; i += 2) {
			String key = String.valueOf(keysAndValues[i]);
			String val = String.valueOf(keysAndValues[i + 1]);
			put(key, val);
		}
	}

	/**
	 * Adds a key/value string pair to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param value the value string for the new param.
	 */
	public void put(String key, String value) {
		if (key != null && value != null) {
			urlParams.put(key, value);
		}
	}

	/**
	 * Adds a file to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param file the file to add.
	 * @throws java.io.FileNotFoundException throws if wrong File argument was
	 *             passed
	 */
	public void put(String key, File file) throws FileNotFoundException {
		put(key, file, null);
	}

	/**
	 * Adds a file to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param file the file to add.
	 * @param contentType the content type of the file, eg. application/json
	 * @throws java.io.FileNotFoundException throws if wrong File argument was
	 *             passed
	 */
	public void put(String key, File file, String contentType) throws FileNotFoundException {
		if (file == null || !file.exists()) {
			throw new FileNotFoundException();
		}
		if (key != null) {
			fileParams.put(key, new FileWrapper(file, contentType));
		}
	}

	/**
	 * Adds an input stream to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param stream the input stream to add.
	 */
	public void put(String key, InputStream stream) {
		put(key, stream, null);
	}

	/**
	 * Adds an input stream to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param stream the input stream to add.
	 * @param name the name of the stream.
	 */
	public void put(String key, InputStream stream, String name) {
		put(key, stream, name, null);
	}

	/**
	 * Adds an input stream to the request.
	 * 
	 * @param key the key name for the new param.
	 * @param stream the input stream to add.
	 * @param name the name of the stream.
	 * @param contentType the content type of the file, eg. application/json
	 */
	public void put(String key, InputStream stream, String name, String contentType) {
		if (key != null && stream != null) {
			streamParams.put(key, new StreamWrapper(stream, name, contentType));
		}
	}

	/**
	 * Adds param with non-string value (e.g. Map, List, Set).
	 * 
	 * @param key the key name for the new param.
	 * @param value the non-string value object for the new param.
	 */
	public void put(String key, Object value) {
		if (key != null && value != null) {
			urlParamsWithObjects.put(key, value);
		}
	}

	/**
	 * Adds string value to param which can have more than one value.
	 * 
	 * @param key the key name for the param, either existing or new.
	 * @param value the value string for the new param.
	 */
	public void add(String key, String value) {
		if (key != null && value != null) {
			Object params = urlParamsWithObjects.get(key);
			if (params == null) {
				// Backward compatible, which will result in "k=v1&k=v2&k=v3"
				params = new HashSet<String>();
				this.put(key, params);
			}
			if (params instanceof List) {
				((List<Object>) params).add(value);
			} else if (params instanceof Set) {
				((Set<Object>) params).add(value);
			}
		}
	}

	/**
	 * Removes a parameter from the request.
	 * 
	 * @param key the key name for the parameter to remove.
	 */
	public void remove(String key) {
		urlParams.remove(key);
		streamParams.remove(key);
		fileParams.remove(key);
		urlParamsWithObjects.remove(key);
	}

	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (ConcurrentHashMap.Entry<String, String> entry : urlParams.entrySet()) {
			if (result.length() > 0)
				result.append("&");

			result.append(entry.getKey());
			result.append("=");
			result.append(entry.getValue());
		}

		for (ConcurrentHashMap.Entry<String, StreamWrapper> entry : streamParams.entrySet()) {
			if (result.length() > 0)
				result.append("&");

			result.append(entry.getKey());
			result.append("=");
			result.append("STREAM");
		}

		for (ConcurrentHashMap.Entry<String, FileWrapper> entry : fileParams.entrySet()) {
			if (result.length() > 0)
				result.append("&");

			result.append(entry.getKey());
			result.append("=");
			result.append("FILE");
		}

		List<BasicNameValuePair> params = getParamsList(null, urlParamsWithObjects);
		for (BasicNameValuePair kv : params) {
			if (result.length() > 0)
				result.append("&");

			result.append(kv.getName());
			result.append("=");
			result.append(kv.getValue());
		}

		return result.toString();
	}

	public void setHttpEntityIsRepeatable(boolean isRepeatable) {
		this.isRepeatable = isRepeatable;
	}

	// public void setUseJsonStreamer(boolean useJsonStreamer) {
	// this.useJsonStreamer = useJsonStreamer;
	// }

	/**
	 * Returns an HttpEntity containing all request parameters
	 * 
	 * @param progressHandler HttpResponseHandler for reporting progress on
	 *            entity submit
	 * @return HttpEntity resulting HttpEntity to be included along with
	 *         {@link org.apache.http.client.methods.HttpEntityEnclosingRequestBase}
	 * @throws java.io.IOException if one of the streams cannot be read
	 */
	public HttpEntity getEntity(ResponseHandlerInterface progressHandler) throws IOException {
		// if (useJsonStreamer) {
		// return createJsonStreamerEntity();
		// } else
		if (streamParams.isEmpty() && fileParams.isEmpty()) {
			return createFormEntity();
		} else {
			return createMultipartEntity(progressHandler);
		}
	}

	// private HttpEntity createJsonStreamerEntity() throws IOException {
	// JsonStreamerEntity entity = new JsonStreamerEntity(!fileParams.isEmpty()
	// || !streamParams.isEmpty());
	// // Add string params
	// for (ConcurrentHashMap.Entry<String, String> entry :
	// urlParams.entrySet()) {
	// entity.addPart(entry.getKey(), entry.getValue());
	// }
	// // Add non-string params
	// for (ConcurrentHashMap.Entry<String, Object> entry :
	// urlParamsWithObjects.entrySet()) {
	// entity.addPart(entry.getKey(), entry.getValue());
	// }
	// // Add file params
	// for (ConcurrentHashMap.Entry<String, FileWrapper> entry :
	// fileParams.entrySet()) {
	// FileWrapper fileWrapper = entry.getValue();
	// entity.addPart(entry.getKey(), new FileInputStream(fileWrapper.file),
	// fileWrapper.file.getName(), fileWrapper.contentType);
	// }
	// // Add stream params
	// for (ConcurrentHashMap.Entry<String, StreamWrapper> entry :
	// streamParams.entrySet()) {
	// StreamWrapper stream = entry.getValue();
	// if (stream.inputStream != null) {
	// entity.addPart(entry.getKey(), stream.inputStream, stream.name,
	// stream.contentType);
	// }
	// }
	// return entity;
	// }

	private HttpEntity createFormEntity() {
		try {
			return new UrlEncodedFormEntity(getParamsList(), contentEncoding);
		} catch (UnsupportedEncodingException e) {
			WCSLogUtil.e("createFormEntity failed");
			return null; // Can happen, if the 'contentEncoding' won't be
			             // HTTP.UTF_8
		}
	}

	private HttpEntity createMultipartEntity(ResponseHandlerInterface progressHandler) throws IOException {
		SimpleMultipartEntity entity = new SimpleMultipartEntity(progressHandler);
		entity.setIsRepeatable(isRepeatable);

		// Add string params
		for (ConcurrentHashMap.Entry<String, String> entry : urlParams.entrySet()) {
			entity.addPart(entry.getKey(), entry.getValue());
		}

		// Add non-string params
		List<BasicNameValuePair> params = getParamsList(null, urlParamsWithObjects);
		for (BasicNameValuePair kv : params) {
			entity.addPart(kv.getName(), kv.getValue());
		}

		// Add stream params
		for (ConcurrentHashMap.Entry<String, StreamWrapper> entry : streamParams.entrySet()) {
			StreamWrapper stream = entry.getValue();
			if (stream.inputStream != null) {
				entity.addPart(entry.getKey(), stream.name, stream.inputStream, stream.contentType);
			}
		}

		// Add file params
		for (ConcurrentHashMap.Entry<String, FileWrapper> entry : fileParams.entrySet()) {
			FileWrapper fileWrapper = entry.getValue();
			entity.addPart(entry.getKey(), fileWrapper.file, fileWrapper.contentType);
		}

		return entity;
	}

	private void init() {
		urlParams = new ConcurrentHashMap<String, String>();
		streamParams = new ConcurrentHashMap<String, StreamWrapper>();
		fileParams = new ConcurrentHashMap<String, FileWrapper>();
		urlParamsWithObjects = new ConcurrentHashMap<String, Object>();
	}

	protected List<BasicNameValuePair> getParamsList() {
		List<BasicNameValuePair> lparams = new LinkedList<BasicNameValuePair>();

		for (ConcurrentHashMap.Entry<String, String> entry : urlParams.entrySet()) {
			lparams.add(new BasicNameValuePair(entry.getKey(), entry.getValue()));
		}

		lparams.addAll(getParamsList(null, urlParamsWithObjects));

		return lparams;
	}

	private List<BasicNameValuePair> getParamsList(String key, Object value) {
		List<BasicNameValuePair> params = new LinkedList<BasicNameValuePair>();
		if (value instanceof Map) {
			Map<String, Object> map = (Map<String, Object>) value;
			List<String> list = new ArrayList<String>(map.keySet());
			// Ensure consistent ordering in query string
			Collections.sort(list);
			for (String nestedKey : list) {
				Object nestedValue = map.get(nestedKey);
				if (nestedValue != null) {
					params.addAll(getParamsList(key == null ? nestedKey : String.format("%s[%s]", key, nestedKey), nestedValue));
				}
			}
		} else if (value instanceof List) {
			List<Object> list = (List<Object>) value;
			for (Object nestedValue : list) {
				params.addAll(getParamsList(String.format("%s[]", key), nestedValue));
			}
		} else if (value instanceof Object[]) {
			Object[] array = (Object[]) value;
			for (Object nestedValue : array) {
				params.addAll(getParamsList(String.format("%s[]", key), nestedValue));
			}
		} else if (value instanceof Set) {
			Set<Object> set = (Set<Object>) value;
			for (Object nestedValue : set) {
				params.addAll(getParamsList(key, nestedValue));
			}
		} else if (value instanceof String) {
			params.add(new BasicNameValuePair(key, (String) value));
		}
		return params;
	}

	protected String getParamString() {
		return URLEncodedUtils.format(getParamsList(), contentEncoding);
	}

	public static class FileWrapper {
		public File file;

		public String contentType;

		public FileWrapper(File file, String contentType) {
			this.file = file;
			this.contentType = contentType;
		}
	}

	public static class StreamWrapper {
		public InputStream inputStream;

		public String name;

		public String contentType;

		public StreamWrapper(InputStream inputStream, String name, String contentType) {
			this.inputStream = inputStream;
			this.name = name;
			this.contentType = contentType;
		}
	}
}
